#include <stdlib.h>
#include <string.h>
#include "treepair.h"

/* crea un treepair vuoto */
treepair *createtreepair(void)
{
	treepair *root = malloc(sizeof(treepair));

	if(!root) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-1);
	}
	root->iroot = root->sroot = NULL;
	return root;
}

/* visita inordine il searchtree (di interi o di stringhe) la cui radice e' p */
/* e applica a ogni nodo la funzione op */
void inorder(treepairnode *p, void (*op)(treepairnode *))
{
	if(p) {
        	inorder(p->left,op);
		(*op)(p);
        	inorder(p->right,op);
	}
}

/* restituisce la radice del searchtree di interi nel treepair di radice p */
/* NULL se p e' NULL */
treepairnode *getintroot(treepair *p)
{
	return p ? p->iroot : NULL;
}

/* restituisce la radice del searchtree di stringhe nel treepair di radice p */
/* NULL se p e' NULL */
treepairnode *getstrroot(treepair *p)
{
	return p ? p->sroot : NULL;
}

/* restituisce il valore intero associato al nodo puntato da p */
/* -1 se p e' NULL o p non punta ad alcun datanode */
int getintdatum(treepairnode *p)
{
	return (p && p->data) ? p->data->intero : -1;
}

/* restituisce la stringa associata al nodo puntato da p */
/* NULL se p e' NULL o p non punta ad alcun datanode */
char *getstrdatum(treepairnode *p)
{
	return (p && p->data) ? p->data->stringa : NULL;
}

/* cerca la chiave k percorrendo l'albero con radice iroot */
treepairnode *isearch(treepairnode *p, int k)
{
	if(!p || k == p->data->intero)
		return p;
	return isearch(k < p->data->intero ? p->left : p->right, k);
}

/* restituisce il puntatore a un nodo nel treeepair di radice p */
/* che abbia valore intero associato uguale a k */
/* NULL se p e' NULL */
treepairnode *intsearch(treepair *p, int k)
{
	/* preleavata la radice p->iroot dell'albero per la ricerca su interi */
	/* innesca la funzione ricorsiva isearch */
	return p ? isearch(p->iroot,k) : NULL;
}

/* cerca la chiave k percorrendo l'albero con radice sroot */
treepairnode *ssearch(treepairnode *p, char *k)
{
	if(!p || strcmp(k,p->data->stringa) == 0)
		return p;
	return ssearch(strcmp(k,p->data->stringa) < 0 ? p->left : p->right, k);
}

/* restituisce il puntatore a un nodo nel treeepair di radice p */
/* che abbia la stringa associata uguale a k */
/* NULL se p e' NULL */
treepairnode *strsearch(treepair *p, char *k)
{
	/* preleavata la radice p->sroot dell'albero per la ricerca su stringhe */
	/* innesca la funzione ricorsiva ssearch */
	return p ? ssearch(p->sroot,k) : NULL;
}

/* restituisce il minimo del searchtree (di interi o di stringhe) di radice p */
treepairnode *treemin(treepairnode *p)  /* si assume p != NULL */
{
	for(;p->left;p = p->left);
	return p;
}

/* restituisce il minimo del searchtree (di interi o di stringhe) di radice p */
treepairnode *treemax(treepairnode *p)  /* si assume p != NULL */
{
	for(;p->right;p = p->right);
	return p;
}

/* restituisce il successore di q nel searchtree cui q appartiene */
treepairnode *treesucc(treepairnode *q)  /* si assume q != NULL */
{
	treepairnode *qq;

	if(q->right)
		return treemin(q->right);
	qq = q->up;	
	while(qq && q == qq->right) {
		q = qq;
		qq = qq->up;
	}
	return qq;
}

/* restituisce il predecessore di q nel searchtree cui q appartiene */
treepairnode *treepred(treepairnode *q)  /* si assume q != NULL */
{
	treepairnode *qq;

	if(q->left)
		return treemax(q->left);
	qq = q->up;	
	while(qq && q == qq->left) {
		q = qq;
		qq = qq->up;
	}
	return qq;
}

/* restituisce 1 se l'intero nel datanode puntato da p e' < dell'intero nel datanode puntato da q */
/* 0 altrimenti */
char intcfr(treepairnode *p, treepairnode *q)  /* si assume p,q != NULL */
{
	return p->data->intero < q->data->intero;
}

/* restituisce 1 se la stringa nel datanode puntato da p e' < della stringa nel datanode puntato da q */
/* 0 altrimenti */
char strcfr(treepairnode *p, treepairnode *q)  /* si assume p,q != NULL */
{
	return strcmp(p->data->stringa, q->data->stringa) < 0 ? 1 : 0;
}

/* inserisce il nodo q nell'albero puntato da p, usando la funzione di confronto cfr */
treepairnode *insertnode
(treepairnode *p, treepairnode *q, char (*cfr)(treepairnode *, treepairnode *))
{
	treepairnode *r = p;
	treepairnode *s = NULL;

	while(r) {
		s = r;
		r = (*cfr)(q,r) ? r->left : r->right;
	}
	q->up = s;
	if(!s)
		return q;
	if((*cfr)(q,s))
		s->left = q;
	else
		s->right = q;
	return p;
}

/* inserisce nel treepair un datanode con valore intero k e stringa s */
void insert(treepair *root, int k, char *s)
{
	treepairnode *qi,*qs;

	/* crea due treepairnode, qi e qs */
	if(!(qi = malloc(sizeof(treepairnode)))) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-1);
	}
	if(!(qs = malloc(sizeof(treepairnode)))) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-1);
	}
	/* crea un datanode e lo assegna a qi->data */
	if(!(qi->data = malloc(sizeof(datanode)))) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-1);
	}
	/* carica k nel campo intero del datanode */
	qi->data->intero = k;
	/* qs->data deve puntare al datanode */
	qs->data = qi->data;
	/* alloca lo spazio necessario e copia s nel campo stringa del datanode */
	if(!(qs->data->stringa = malloc(strlen(s) + 1))) {
		fprintf(stderr, "Errore nell'allocazione della memoria\n");
		exit(-1);
	}
	strcpy(qs->data->stringa,s);
	/* inizializza a NULL i puntatori ai figli di entrambi i treepairnode */
	qi->left = qi->right = qs->left = qs->right = NULL;
	/* stabilisce i legami dal datanode ai due treepairnode */
	qi->data->i = qi;
	qs->data->s = qs;
	/* inserisce qi nell'albero di radice iroot e qs in quello di radice sroot */
	root->iroot = insertnode(root->iroot,qi,intcfr);
	root->sroot = insertnode(root->sroot,qs,strcfr);
}

/* disconnette il nodo q dall'albero di radice p, liberando lo spazio se flag != 0 */
treepairnode *deletenode(treepairnode *p, treepairnode *q, char flag)  /* si assume q != NULL */
{
	treepairnode *r, *s, *t = NULL, *d = q->data->i;

	/* libera il datanode e la stringa contenuta nel datanode, se flag e' settato */
	if(flag) {
		free(q->data->stringa);
		free(q->data);
	}
	if(!q->left || !q->right)
		r = q;
	else
		r = treesucc(q);
	s = r->left ? r->left : r->right;
	if(s)
		s->up = r->up;
	if(!r->up)
		t = s;
	else
		if(r == r->up->left)
			r->up->left = s;
		else
			r->up->right = s;
	/* se stiamo rimuovendo il successore r al posto di q, dobbiamo salvare i dati di r in q */
	/* e ricollegare il nodo che adesso contiene i dati di r, al treepairnode corrispondente nell'altro albero */
	if(r != q) {
		/* se q == d allora q e' nell'albero di radice iroot */
		if(q == d)
			/* ricollega il nodo corrispondente a r nell'albero di radice sroot */
			/* al nodo q che ora contiene i dati di r */
			r->data->i = q; 
		else
			/* ricollega il nodo corrispondente a r nell'albero di radice iroot */
			/* al nodo q che ora contiene i dati di r */
			r->data->s = q;
		/* salva in q i dati di r */
		q->data = r->data;
	}
	free(r);
	return t ? t : (p != r ? p : NULL);
}

/* cancella dal treepair il nodo puntato da q, il datanode associato e il */
/* nodo associato al datanode nell'altro searchtree del treepair */
void delete(treepair *root, treepairnode *q) /* si assume q != NULL */
{
	treepairnode *t;

	/* se q si trova nell'albero di radice iroot */
	if(q == q->data->i) {
		/* assegna a t il nodo corrispondente nell'albero di radice sroot */
		t = q->data->s;
		/* disconnette il nodo q dall'albero di radice iroot */
		root->iroot = deletenode(root->iroot,q,0);
		/* disconnette e cancella i dati associati al nodo t dall'albero di radice sroot */
		root->sroot = deletenode(root->sroot,t,1);
	} else {
		/* assegna a t il nodo corrispondente nell'albero di radice iroot */
		t = q->data->i;
		/* disconnette il nodo q dall'albero di radice sroot */
		root->sroot = deletenode(root->sroot,q,0);
		/* disconnette e cancella i dati associati al nodo t dall'albero di radice iroot */
		root->iroot = deletenode(root->iroot,t,1);
	}
}



